class Layer(object):

    def __init__(self):
        self.prevlayer = None
        self.nextlayer = None

    def forepropagation(self):
        pass

    def backpropagation(self):
        pass

    def initialization(self):
        pass
